@@ -80,6 +80,12 @@ xtab_export_read(void)
 	return xtab_read(_PATH_ETAB, 1);
 }
 
+/*
+ * mountd now keeps an open fd for the etab at all times to make sure that the
+ * inode number changes when the xtab_export_write is done. If you change the
+ * routine below such that the files are edited in place, then you'll need to
+ * fix the auth_reload logic as well...
+ */
 static int
 xtab_write(char *xtab, char *xtabtmp, int is_export)
 {
@@ -102,10 +108,7 @@ xtab_write(char *xtab, char *xtabtmp, int is_export)
 
 			/* write out the export entry using the FQDN */
 			xe = exp->m_export;
-			strncpy(xe.e_hostname,
-				exp->m_client->m_hostname,
-				sizeof (xe.e_hostname) - 1);
-			xe.e_hostname[sizeof (xe.e_hostname) - 1] = '\0';
+			xe.e_hostname = exp->m_client->m_hostname;
 			putexportent(&xe);
 		}
 	}
@@ -140,9 +143,7 @@ xtab_append(nfs_export *exp)
 		return;
 	setexportent(_PATH_XTAB, "a");
 	xe = exp->m_export;
-	strncpy(xe.e_hostname, exp->m_client->m_hostname,
-	       sizeof (xe.e_hostname) - 1);
-	xe.e_hostname[sizeof (xe.e_hostname) - 1] = '\0';
+	xe.e_hostname = exp->m_client->m_hostname;
 	putexportent(&xe);
 	endexportent();
 	xfunlock(lockid);
