@@ -45,7 +45,6 @@ extern char *  SM_STAT_PATH;
  * Function prototypes.
  */
 extern void	change_state(void);
-extern void	do_regist(u_long, void (*)());
 extern void	my_svc_run(void);
 extern void	notify_hosts(void);
 extern void	shuffle_dirs(void);
@@ -54,6 +53,7 @@ extern int	process_reply(FD_SET_TYPE *);
 extern char *	xstrdup(const char *);
 extern void *	xmalloc(size_t);
 extern void	xunlink (char *, char *, short int);
+extern void	load_state(void);
 
 /*
  * Host status structure and macros.
@@ -82,12 +82,11 @@ extern int run_mode;
  * that just came back up, for ex, when failing over a HA service to
  * another host.... */
 #define STATIC_HOSTNAME 8	/* Always use the hostname set by -n */
-
+#define	MODE_NO_NOTIFY	16	/* Don't notify peers of a reboot */
 /*
  * Program name and version pointers -- See statd.c for the reasoning
  * as to why they're global.
  */
 extern char *name_p;		/* program basename */
-extern char *version_p;		/* program version */
+extern const char *version_p;	/* program version */
 
-extern int		re_notify; /* time to re-read notify list */
